var _ = 		require('underscore'),
	util =      require('util'),
	config = 	require('./config'),
	logger =    require('../logger/index'),

	AbstractWarn = function (errType, code, constr) {
		Error.captureStackTrace(this, constr || this)
		this.message = config.messages['EN'][errType][code] + ' (' + code + ')'
		this.code = code;
		if (code === 350 || code === 351) {
			logger.error(this.stack);
		} else {
			logger.warn (this.stack);	
		}
	},

	AbstractError = function (errType, code, constr) {
		Error.captureStackTrace(this, constr || this)
		this.message = config.messages['EN'][errType][code] + ' (' + code + ')'
		this.code = code;
		logger.error (this.stack);
	};

util.inherits(AbstractError, Error)
AbstractError.prototype.name = 'Abstract Error'

util.inherits(AbstractWarn, Error)
AbstractWarn.prototype.name = 'Abstract Warn'



var ApplicationError = function (code) { ApplicationError.super_.call(this, 'app', code, this.constructor) };
util.inherits(ApplicationError, AbstractWarn);
ApplicationError.prototype.name = 'Application Error';

var DatabaseError = function (code) { DatabaseError.super_.call(this, 'db', code, this.constructor) };
util.inherits(DatabaseError, AbstractError);
DatabaseError.prototype.name = 'Database Error';

var ServerError = function (code) { ServerError.super_.call(this, 'srv', code, this.constructor) };
util.inherits(ServerError, AbstractError);
ServerError.prototype.name = 'Server Error';






var AWSError = function (code) { AWSError.super_.call(this, 'aws', config.awsErrorTranslate[code] || 800, this.constructor) };
util.inherits(AWSError, AbstractWarn);
AWSError.prototype.name = 'AWS Error';

var FBError = function (code) { FBError.super_.call(this, 'fb', config.fbErrorTranslate[code] || 950, this.constructor) };
util.inherits(FBError, AbstractWarn);
FBError.prototype.name = 'FB Error';

var PaypalError = function (code) { PaypalError.super_.call(this, 'ppl', code || 1000, this.constructor) };
util.inherits(PaypalError, AbstractWarn);
PaypalError.prototype.name = 'Paypal Error';

var StripeError = function (code) { StripeError.super_.call(this, 'stp', config.stpErrorTranslate[code] || 1200, this.constructor) };
util.inherits(StripeError, AbstractWarn);
StripeError.prototype.name = 'Stripe Error';

module.exports = {
	db: 	DatabaseError,
	srv: 	ServerError,
	app: 	ApplicationError,
	aws: 	AWSError,
	fb: 	FBError,
	ppl: 	PaypalError,
	stp: 	StripeError,
	init: 	options => { if (options && _.isObject(options)) _.extend(config, options); }
}