"use strict";

let Promise             =   require ('bluebird'),
    _                   =   require ('underscore'),
    fs                  =   require ('fs'),
    util                =   require ('util'),

    passport            =   require ('passport'),
    bodyParser          =   require ('body-parser'),
    // cookie              =   require ('express/node_modules/cookie'),
    cookie              =   require ('cookie'),
    cookieParser        =   require ('cookie-parser'),
    // formidable          =   require ('formidable'),
    textParser          =   bodyParser.text(),

    utilities           =   require ('../utilities/index'),
    validator           =   require ('../validator/index'),
    sessions            =   require ('../sessions/index'),
    tmpFolder           =   __dirname + '/../.tmp/',
    io                  =   null,
    controllers         =   { },
    env                 =   process.env.NODE_ENV || 'development';


function getUserAgent(req) {
    if (req.headers['x-ucbrowser-ua']) {  //special case of UC Browser
        return 'User-agent: ' + (req.headers['x-ucbrowser-ua'] || '');
    } else {
        return 'User-agent: ' + (req.headers['user-agent'] || '');
    }
}

// function validateAndSanitize (args, api, session, callback) {
function validateAndSanitize (data) {
    utilities.logger.info('framework/interfaces validateAndSanitize');

    return new Promise (function (resolve, reject) {


        let api = data.api,
            args = data.req.method === 'POST' ? data.req.body : data.req.query,
            session = data.req.session;

        if (data.command.skipValidate) {
            data.req.data = args;
            return resolve (data);
        }

        // check for valid session if required
        if (!data.command.authBypass) {
            utilities.logger.info ('CHECKAUTH - ' + api.section + ' ' + api.command);

            if (!session || !session.passport || !session.passport.user) {
                utilities.logger.info ('NOSESSION - ' + api.section + ' ' + api.command);
                return reject( new utilities.error.app(405) );
            }

            if (!session.passport.user.active) {
                //utilities.logger.info ('ACCOUNT DISABLED - ' + api.section + ' ' + api.command);
                sessions.removeSessionKeys ({_person: session.passport.user._id});
                return reject( new utilities.error.app(408) );
            }
        }

        // check for valid arguments
        let requiredProperties = data.command.args.required || [],
            optionalProperties = data.command.args.optional || [],
            allProperties = _.union (requiredProperties, optionalProperties),
            validArgs = allProperties.length ? ( _.pick (args, allProperties) ) : {};

        if ( requiredProperties.length && validator.missingProperties ( validArgs, requiredProperties ) ) return reject ( new utilities.error.app(350) );

        allProperties = _.intersection(_.keys(validArgs), allProperties) || []; // filter allProperties for only ones in the argument list

        return validator.isValidPromise(validArgs, allProperties)
            .then(validator.sanitize)
            .then(cleanData => {
                data.req.data = cleanData;
                resolve (data);
                return null;
            })
            .catch(reject);
    });
};

function validateAndSanitizePromise (data) {
    utilities.logger.info('framework/interfaces validateAndSanitize');

    return new Promise (function (resolve, reject) {


        let api = data.api,
            args = data.req.method === 'POST' ? data.req.body : data.req.query,
            session = data.req.session;

        if (data.command.skipValidate) {
            data.req.data = args;
            return resolve (data);
        }

        // check for valid session if required
        if (!data.command.authBypass) {
            utilities.logger.info ('CHECKAUTH - ' + api.section + ' ' + api.command);

            if (!session || !session.passport || !session.passport.user) {
                utilities.logger.info ('NOSESSION - ' + api.section + ' ' + api.command);
                return reject(new utilities.error.app(405));
            }

            if (!session.passport.user.active) {
                //utilities.logger.info ('ACCOUNT DISABLED - ' + api.section + ' ' + api.command);
                sessions.removeSessionKeys ({_person: session.passport.user._id});
                return reject(new utilities.error.app(408));
            }
        }

        // check for valid arguments
        let requiredProperties = data.command.args.required || [],
            optionalProperties = data.command.args.optional || [],
            allProperties = _.union (requiredProperties, optionalProperties),
            validArgs = allProperties.length ? ( _.pick (args, allProperties) ) : {};

        if ( requiredProperties.length && validator.missingProperties ( validArgs, requiredProperties ) ) {
            return reject(new utilities.error.app(350));
        }

        allProperties = _.intersection(_.keys(validArgs), allProperties) || []; // filter allProperties for only ones in the argument list

        return validator.isValidPromise(validArgs, allProperties)
            .then(validator.sanitize)
            .then(cleanData => {
                data.args = cleanData;
                resolve (data);
                return null;
            })
            .catch(e => {
                data.e = e;
                reject (data.e);
                return null;
            });
    });
};

function obeyCommandREST (data) {
        controllers[data.api.version][data.api.section][data.api.command] (data.req, data.res, function (e, result, user) {
            if (e) {
                utilities.logger.info('REST ERROR: ' + data.api.section + ' ' + data.api.command + ' * error *: ' + util.inspect(e));
                utilities.logger.error(getUserAgent(data.req));
                if (result) {
                    data.res.send( _.extend ( { error: e.code || -1 }, result) );
                } else {
                    data.res.send( { error: e.code || -1 } );
                }
            } else {
                utilities.logger.info('REST: ' + data.api.section + ' ' + data.api.command + ' success.');
                if (user) sessions.updateSessions (user);
                data.res.send( _.extend ({ error: 0 }, result) );
            }
        });
        return Promise.resolve();
}


function getRESTFunction (api) {
    if (env === 'production') {
        return (req, res) => validateAndSanitize ({api:api,command:controllers[api.version][api.section].api[api.command], req:req, res:res, result: {}, e: null})
            .then(obeyCommandREST)
            .catch(e => {
                utilities.logger.error(getUserAgent(req));
                res.send({error: e.code || -1});
            });
    } else {
        return (req, res) => {
            utilities.logger.info ('COOKIES: ' + JSON.stringify (req.cookies) );
            utilities.logger.info ('REST: ' + api.section + ' ' + api.command );
            utilities.logger.info ('Data: ' + JSON.stringify(req.method === 'POST' ? req.body : req.query) );

            validateAndSanitize ({api:api,command:controllers[api.version][api.section].api[api.command], req:req, res:res, result: {}, e: null})
                .then(obeyCommandREST)
                .catch(e => {
                    utilities.logger.error(getUserAgent(req));
                    res.send({error: e.code || -1});
                });
        };
    }

}

function getRESTFunctionPromise (api) {
    return  ( req, res ) => {
        utilities.logger.info ('COOKIES: ' + JSON.stringify (req.cookies) );
        utilities.logger.info ('REST: ' + api.section + ' ' + api.command );
        utilities.logger.info ('Data: ' + JSON.stringify(req.method === 'POST' ? req.body : req.query) );

        let data = {
            api: api,
            args: {}, // clean args passed from client
            command: controllers[api.version][api.section].apiPromise[api.command], // command properties
            req: req,
            res: res,
            result: null,
            temp: {},
            user: null,
        };

        return validateAndSanitizePromise (data)
            .then(controllers[api.version][api.section][api.command])
            .then(result => {
                utilities.logger.info('REST: ' + api.section + ' ' + api.command + ' success.');
                if (data.user) sessions.updateSessions (data.user);
                return res.send( _.extend ({ error: 0 }, result || {}) );
            })
            .catch(e => {
                utilities.logger.info('REST ERROR: ' + api.section + ' ' + api.command + ' * error *: ' + util.inspect(e));
                utilities.logger.error(getUserAgent(req));
                if (data.result) {
                    return res.send( _.extend ( { error: e.code || -1 }, data.result) );
                } else {
                    return res.send( { error: e.code || -1 } );
                }
            });
    };
}


//// Public --------------------------------------------------------------------------------------


exports.initREST = function (server, options) {
    utilities.logger.info('framework/interfaces initREST');

    // listen for each rest command
    _.each ( controllers, (sections, version) =>
        _.each (sections, (commands, section) => {
            _.each (commands.api, (commandsDetail, command) =>
                _.each (commandsDetail.methods, method => {
                        if (controllers[version][section].api[command].parseText) server[method] ( '/'+version+'/'+section+'/'+command, textParser, getRESTFunction({version:version, section:section, command:command}));
                        else server[method] ( '/'+version+'/'+section+'/'+command, getRESTFunction({version:version, section:section, command:command}))
                    }
                )
            )

            _.each (commands.apiPromise, (commandsDetail, command) =>
                _.each (commandsDetail.methods, method => {
                        if (controllers[version][section].apiPromise[command].parseText) server[method] ( '/'+version+'/'+section+'/'+command, textParser, getRESTFunctionPromise({version:version, section:section, command:command}));
                        else server[method] ( '/'+version+'/'+section+'/'+command, getRESTFunctionPromise({version:version, section:section, command:command}));
                    }
                )
            )
        })
    );
}

/* ----------------------------------------------------------------------------------------------------------------------------------*/

exports.initWebsockets = function ( serverInstance, options ) {
    utilities.logger.info('framework/interfaces initWebsockets');

    io = require('socket.io')(serverInstance, options.socketIO);

    let getWebsocketsFunction = function (api) {
        return function (data, fileData) {

            let socket = this,
                currentCommand = controllers[api.version][api.section].api[api.command],
                path = '/'+api.version+'/'+api.section+'/'+api.command,

                returnFunc = function(e, result, user){

                    if (user) {
                        socket.request.session[passport._key] = {};
                        socket.request.session[passport._key][passport._userProperty] = user;
                        options.session.store.set(socket.request.session.id, socket.request.session, function(e, sessionInfo){ if (e) utl.logger.warn ('Error saving session data'); });
                    }

                    if (e) {
                        utilities.logger.info('WEBSOCKETS: ' + path + ' * error *: ' + e);
                        socket.emit(path, { error: e.code || -1 } );
                    } else {
                        utilities.logger.info('WEBSOCKETS: ' + path + ' success.');
                        socket.emit(path, _.extend ( { error: 0 }, result) );
                    }
                },

                goValidate = function () {
                    validateAndSanitize (data, api, socket.request.session)
                        .then(cleanData => {
                            if (!cleanData.username && cleanData.email) cleanData.username = cleanData.email;
                            let req = _.extend ( {}, {session: socket.request.session || {}, body: socket.request.body || {}, query:socket.request.query || {}, headers: socket.request.headers || {} }, {body:cleanData, query:cleanData, data: cleanData, url:'/socket.io'} );
                                res = {};
                            controllers[api.version][api.section][api.command] (req, res, returnFunc);
                        })
                        .catch(e => socket.emit('/'+api.version+'/'+api.section+'/'+api.command, { error: e.code || -1 } ) );
                };

            //utilities.logger.info('SOCKET SESSION: ' + JSON.stringify(socket.request.session));

            //if ( currentCommand.fileUploadAllowed && data.files && data.files.length && buffer && buffer.length) {
            if ( currentCommand.fileUploadAllowed ) {
                if ( !data.files || !data.files.length ) utilities.logger.info('WEBSOCKETS: files parameter not found or empty' );
                else {
                    if ( !fileData || !fileData.length) utilities.logger.info('WEBSOCKETS: fileData parameter not found or empty' );
                    else {
                        let afterFileUpload = _.after(data.files.length, goValidate);
                        _.each (data.files, function (file, index) {
                            let fileName = tmpFolder + Date.now() + '_' + file;
                            fs.open(fileName, 'a', '0755', function (e, fd) {
                                if (e) utilities.logger.warn('WEBSOCKETS: unable to write ' + fileName );
                                else {
                                    fs.write(fd, fileData[index], null, 'Binary', function(e, written, buff) {
                                        if (e) utilities.logger.warn('WEBSOCKETS: error writing ' + fileName );
                                        fs.close(fd, function() {
                                            utilities.logger.info (fileName + ' saved successfully!');
                                            data.files [index] = { path: fileName, size: fs.statSync(fileName).size, mimeType: 'image/jpeg' };
                                            afterFileUpload();
                                        });
                                    });
                                }
                            });
                        });
                    }
                }
            } else goValidate();
        };
    };

    //io.of('/shufl')
    io.use(function(socket, next) {
        let cookieHeader = socket.request.headers.cookie, session = null, id = null;

        if (!cookieHeader) return next('No cookie transmitted.');
        cookieHeader = cookie.parse(cookieHeader);

        utilities.logger.info(cookieHeader);

        if (cookieHeader[options.session.secret]) {
            id = cookieParser.signedCookie (cookieHeader[options.session.secret], options.session.secret);

            if (cookieHeader[options.session.secret] == id || id === false) {
                utilities.logger.warn('Cookie is invalid');
                return next('Cookie is invalid');
            }
            options.session.store.get(id, function(e, sessionInfo){
                if (sessionInfo) socket.request.session = sessionInfo;
                else socket.request.session = {};
                socket.request.session.id = id;
                utilities.logger.info('session id: ' + socket.request.session.id);
                utilities.logger.info('session: ' + JSON.stringify(socket.request.session));
                return next();
            });
        } else return next('No cookie transmitted.');
    })
    .on('connection', function (socket) {
        utilities.logger.info('Connected');
        socket.on ('error',         function (e) { utilities.logger.warn('Connection error'); utilities.logger.warn(e); });
        socket.on ('terminate',     function () { socket.disconnect(); utilities.logger.info('Client terminated connection'); });
        socket.on ('disconnect',    () => utilities.logger.warn('Disconnected') );

        // custom functions
        _.each ( controllers, function(sections, version){
            _.each (sections, function(commands, section){
                _.each (commands.api, function(commandsDetail, command){
                    socket.on ( '/'+version+'/'+section+'/'+command, getWebsocketsFunction ( { version:version, section:section, command:command } ) );
                });
            });
        });
    });
}

/* ----------------------------------------------------------------------------------------------------------------------------------*/

exports.setControllers = function ( customControllers ) {
    if (customControllers && _.isObject(customControllers)) _.extend (controllers, customControllers);
    exports.currentControllers  = controllers['1.0']
};

/* ----------------------------------------------------------------------------------------------------------------------------------*/

exports.controllers         =  controllers;






/*function deleteLocalFile (fullPath) {
    fs.unlink (fullPath, function (e) {
        if (e) {
            utilities.logger.error(e);
            new utilities.error.srv(203);
        }
        else utilities.logger.info ('Deleted: ' + fullPath);
    });
}*/

/*function handleMultipartRest (req, res, api) {
    let form = new formidable.IncomingForm();
    form.uploadDir = tmpFolder;
    form.multiples = true;

    form.parse(req, function(e, data, files) {
        utilities.logger.info ('Form Data: ' + JSON.stringify(data) );
        if (e) {
            utilities.logger.error('REST: ' + api.section + ' ' + api.command + ' * error *: ' + util.inspect(e));
            _.each(files, file => deleteLocalFile (file.path) );
            return res.send( { error: (new utilities.error.app(409)).code || -1 } );
        } else {
            let fileArray = [];
            _.each(files, file => fileArray.push( { size: file.size, path:file.path, mimeType:file.type, type:file.type } ) );

            //multipart data is usually url encoded. Decode strings here.
            _.each (data, function (value, prop) {
                if ( _.isString (value) ) {
                    try {
                        data[prop] = decodeURIComponent(value);
                    }
                    catch (e) {
                        utilities.logger.warn ("Decode Component Failed!");
                        utilities.logger.warn (e);
                        utilities.logger.warn (value);
                    }
                }
            });

            return goValidateREST ({args:_.extend ({}, data, { files:fileArray }), api:api, req:req, res:res});
        }
    });
}*/

/*function getRESTFunction (api) {
    if (env === 'production') return ( req, res ) => validateAndSanitize ({api:api, req:req, res:res})
        .then(obeyCommandREST)
        .catch(e => res.send( { error: e.code || -1 } ) );
    else return  ( req, res ) => {
            utilities.logger.info ('COOKIES: ' + JSON.stringify (req.cookies) );
            utilities.logger.info ('REST: ' + api.section + ' ' + api.command );
            utilities.logger.info ('Data: ' + JSON.stringify(req.method === 'POST' ? req.body : req.query) );

            if ( req.is('multipart/form-data')  ) {
                if (controllers[api.version][api.section].api[api.command].fileUploadAllowed) handleMultipartRest (req, res, api);
                else return res.send( { error: 200} );
            } else return validateAndSanitize ({api:api, req:req, res:res})
                .then(obeyCommandREST)
                .catch(e => res.send( { error: e.code || -1 } ) );
        };
}*/